package com.ericsson.p2psip.m2m;

import java.io.BufferedReader;
import java.io.Console;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.SocketException;
import java.net.URI;
import java.net.URISyntaxException;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.Random;

import com.ericsson.coap.coap.MediaTypeRegistry;
import com.ericsson.coap.coap.Option;
import com.ericsson.coap.coap.OptionNumberRegistry;
import com.ericsson.coap.coap.RemoteResource;
import com.ericsson.coap.coap.Request;
import com.ericsson.coap.coap.Resource;
import com.ericsson.coap.coap.Response;
import com.ericsson.coap.endpoint.Endpoint;
import com.ericsson.coap.example.SampleClient;
import com.ericsson.coap.example.SampleServer;
import com.ericsson.p2psip.util.IPEndpoint;

/**
 * This is the application for the Gumstix Overo Earth in:
 * 
 * Gumstix Overlay Gumstix (Overo Earth) (Pinto TH) │ p2pGetIP(coap) │ │ Input
 * via ssh │──────────────────────>│ │ │<─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ │ │ │ │ │ │ │ │ │
 * CoAPMessage() │ │ │──────────────────────────────────────>│ │ -Target's CoAP
 * name │ │ │ -"Button pressed" │ │ Print message on screen │ │ │ │ │ │
 * 
 * @author jaime
 * 
 */

public class M2MExampleRmi4Measurements {

	public long timestamp;

	public static void main(String[] args) throws InterruptedException,
			RemoteException {
		/* Start the nodes */
		if (args.length < 1) {
			System.err.println("Wrong number of arguments. Usage: java -jar "
					+ "m2mce.jar -c config_file.txt ");
			System.exit(1);
		}
		int i = 0;
		if (args[i++].equals("-c")) {
			P2P4M2MCE.CONFIG_FILE = args[i].toString();
		} else {
			System.err.println("ERROR: -c is missing");
			return;
		}

		IM2M ce = null;
		try {
			ce = new P2P4M2MCE();
		} catch (RemoteException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		// Create and install a security manager
		if (System.getSecurityManager() == null) {
			System.setSecurityManager(new RMISecurityManager());
			System.out.println("Security manager installed.");
		} else {
			System.out.println("Security manager already exists.");
		}

		Registry registry = null;
		try { // special exception handler for registry creation
			registry = LocateRegistry.createRegistry(1099);
			System.out.println("java RMI registry created.");
		} catch (RemoteException e) {
			// do nothing, error means registry already exists
			System.out.println("java RMI registry already exists.");
		}

		String rmiName = "M2MCE";
		try {
			// Naming.bind(rmiName, ce);
			registry.rebind(rmiName, ce);
		} catch (Exception e) {
			System.err.println("M2MCE RMI server binding error");
			e.printStackTrace();
			return;
		}
		System.out.println("M2MCE RMI server started with name " + rmiName);
		System.out.println("Starting test");

		/* A PN or WN joins */
		String name = ce.getBookData().getPnode().getPn().getName();
		ce.join(name, true);
		Thread.sleep(1000);

		UpdateList ul1 = ce.getUpdates();
		System.out.println("Printing Update List: ");
		ul1.print();
		Thread.sleep(1000);
		Book b1 = ce.getList();
		System.out.println("Printing Book: ");
		b1.print();
		Thread.sleep(1000);

		try {

			Endpoint server = new SampleServer();

			System.out.printf("SampleServer listening at port %d.\n",
					server.port());

		} catch (SocketException e) {

			System.err.printf("Failed to create SampleServer: %s\n",
					e.getMessage());

		}
		for (int msgsize = 1; msgsize <= 180; msgsize++) {

			String node = "jaime/ln0";
			String msg = generateString(
					new Random(),
					"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789abcdefghijklmnopqrstuvwxyz",
					msgsize);
			long startTime = System.currentTimeMillis();
			String ip = ce.getIP(node).getIpAddress();
			long stopTime = System.currentTimeMillis();
			long lookup = stopTime - startTime;
			System.out.println("Message size: " + msgsize);
			System.out.println("Lookup time: " + lookup);
			String[] coapinput = new String[] { "POST",
					"coap://" + ip + ":5683/proxy", "jaime/ln0:" + msg };
			coapmain(coapinput);
		}

	}

	public static String generateString(Random rng, String characters,
			int length) {
		char[] text = new char[length];
		for (int i = 0; i < length; i++) {
			text[i] = characters.charAt(rng.nextInt(characters.length()));
		}
		return new String(text);
	}

	public static void coapmain(String[] args) {

		// initialize parameters
		String method = null;
		String uri = null;
		String payload = null;
		boolean loop = false;

		// display help if no parameters specified
		if (args.length == 0) {
			SampleClient.printInfo();
			return;
		}

		// input parameters
		int idx = 0;
		for (String arg : args) {
			if (arg.startsWith("-")) {
				if (arg.equals("-l")) {
					loop = true;
				} else {
					System.out.println("Unrecognized option: " + arg);
				}
			} else {
				switch (idx) {
				case SampleClient.IDX_METHOD:
					method = arg.toUpperCase();
					break;
				case SampleClient.IDX_URI:
					uri = arg;
					break;
				case SampleClient.IDX_PAYLOAD:
					payload = arg;
					break;
				default:
					System.out.println("Unexpected argument: " + arg);
				}
				++idx;
			}
		}

		// create request according to specified method
		if (method == null) {
			System.err.println("Method not specified");
			return;
		}
		Request request = SampleClient.newRequest(method);
		if (request == null) {
			System.err.println("Unknown method: " + method);
			return;
		}

		if (method.equals("OBSERVE")) {
			request.setOption(new Option(60, OptionNumberRegistry.OBSERVE));
			loop = true;
		}

		// set request URI
		if (uri == null) {
			System.err.println("URI not specified");
		}
		if (method.equals("DISCOVER")
				&& !uri.endsWith(SampleClient.DISCOVERY_RESOURCE)) {
			uri = uri + SampleClient.DISCOVERY_RESOURCE;
		}
		try {
			request.setURI(new URI(uri));
		} catch (URISyntaxException e) {
			System.err.println("Failed to parse URI: " + e.getMessage());
			return;
		}
		// set request payload
		request.setPayload(payload);

		// enable response queue in order to use blocking I/O
		request.enableResponseQueue(true);

		// execute request
		try {
			request.execute();
		} catch (IOException e) {
			System.err.println("Failed to execute request: " + e.getMessage());
			return;
		}

		// loop for receiving multiple responses
		do {

			// receive response

			System.out.println("Receiving response...");
			Response response = null;
			try {
				response = request.receiveResponse();

				// check for indirect response
				if (response != null && response.isEmptyACK()) {
					response.log();
					System.out
							.println("Request acknowledged, waiting for separate response...");

					response = request.receiveResponse();
				}

			} catch (InterruptedException e) {
				System.err.println("Failed to receive response: "
						+ e.getMessage());
				return;
			}

			// output response

			if (response != null) {

				response.log();
				System.out
						.println("Round Trip Time (ms): " + response.getRTT());

				// check of response contains resources
				if (response.hasFormat(MediaTypeRegistry.LINK_FORMAT)) {

					String linkFormat = response.getPayloadString();

					// create resource three from link format
					Resource root = RemoteResource.newRoot(linkFormat);
					if (root != null) {

						// output discovered resources
						System.out.println("\nDiscovered resources:");
						root.log();

					} else {
						System.err.println("Failed to parse link format");
					}
				} else {

					// check if link format was expected by client
					if (method.equals("DISCOVER")) {
						System.out
								.println("Server error: Link format not specified");
					}
				}

			} else {

				// no response received
				// calculate time elapsed
				long elapsed = System.currentTimeMillis()
						- request.getTimestamp();

				System.out.println("Request timed out (ms): " + elapsed);
				break;
			}

		} while (loop);

		// finish
		System.out.println();
	}

}
