package com.ericsson.p2psip.m2m;

import java.io.IOException;
import java.io.Serializable;

import com.ericsson.jiam.Configuration;
import com.ericsson.jiam.Logger;
import com.ericsson.jiam.MIDlet;
import com.ericsson.jiam.sip.SipConstants;
import com.ericsson.jiam.sip.SipICEConnection;
import com.ericsson.p2psip.ConnectionManager;
import com.ericsson.p2psip.IChordCallbackReceiver;
import com.ericsson.p2psip.ILookupResultReceiver;
import com.ericsson.p2psip.IPublishResultReceiver;
import com.ericsson.p2psip.IRemoveResultReceiver;
import com.ericsson.p2psip.P2ppConstants;
import com.ericsson.p2psip.chord.Chord;
import com.ericsson.p2psip.ice.IIncomingAppAttachHandler;
import com.ericsson.p2psip.logger.P2pSipLogger;
import com.ericsson.p2psip.measurement.FullMeasurementData;
import com.ericsson.p2psip.util.ConfigData;
import com.ericsson.p2psip.util.DHTEvent;
import com.ericsson.p2psip.util.IPEndpoint;
import com.ericsson.p2psip.util.LookupResult;
import com.ericsson.p2psip.util.OverlayData;
import com.ericsson.p2psip.util.ResourceRecord;
import com.ericsson.p2psip.util.SecureHashAlgorithm;

/**
 * Implements IP2P4M2M, i.e., acts as an interface between the P2PSIP prototype
 * and an application
 * @author ejoumae
 *
 */
public class P2P4M2MEngine implements IP2P4M2M, IChordCallbackReceiver, 
		IIncomingAppAttachHandler, ILookupResultReceiver, 
		IPublishResultReceiver, IRemoveResultReceiver, Serializable
{

	/**
	 * 
	 */
	private static final long serialVersionUID = -3973585234044810968L;

	/**
	 * The DHT
	 */
	private Chord dht;
	
	/**
	 * Callback listener listening to callbacks from P2P4M2MEngine
	 */
	private IP2P4M2MCallback callbackListener;
	
	/**
	 * The logger
	 */
	private P2pSipLogger logger;
	
	/**
	 * For storing data used in experiments
	 */
	private FullMeasurementData measurementData;

	public String config_file;

//	/**
//	 * The configuration file
//	 */
//	public static final String CONFIG_FILE = "config.txt";
	
	/**
	 * A string used to identify the P2P overlay
	 */
	public static final String OVERLAY_NAME = "p2p4m2m";
	
	
	/**
	 * Create a P2P4M2MEngine instance
	 * @param callbackListener is the callback listener listening to callbacks
	 * from P2P4M2MEngine
	 */
	public P2P4M2MEngine(IP2P4M2MCallback callbackListener) {
		this.config_file = "null";
		this.callbackListener = callbackListener;
		this.measurementData = new FullMeasurementData();
		this.logger = P2pSipLogger.getInstance("P2P4M2MEngine");
		ConfigData cfgData = new ConfigData();
		P2P4M2MUtils.setConfiguration(cfgData);
    	byte[] overlayNameBytes = OVERLAY_NAME.getBytes();
    	OverlayData od = new OverlayData(overlayNameBytes, 
    			overlayNameBytes.length, P2ppConstants.SHA_1, 
    			P2ppConstants.SHA_1_HASH_LENGTH, P2ppConstants.CHORD);
    	
    	int sipPort = Configuration.getProperty("sipServerPortPublic", 
    			SipConstants.DEFAULT_SIP_PORT);
		IPEndpoint sipIP = new IPEndpoint(cfgData.getProperty(
				ConfigData.LOCAL_IP_STRING), sipPort, 
				P2ppConstants.TRANSPORT_UDP);

		this.dht = new Chord(cfgData, "m2m", od, this, sipIP, 
				this.measurementData, this);

		
        try {
        	int localServerPort = cfgData.getIntProperty(ConfigData.LOCAL_PORT);
        	String transport = cfgData.getProperty(
        			ConfigData.TRANSPORT_PROTOCOL);
        	logger.debug("P2P4M2MEngine.initialize( ): Opening the local server"
        			+ " port (" + localServerPort + ").");
        	ConnectionManager cm = dht.getData().getConnectionManager();
        	cm.openServerPort(dht.getCore(), localServerPort, transport);
        	
        }
        catch(IOException e) {
        	Logger.log("P2P4M2MEngine.initialize( ): An exception was caught "
        			+ "while opening P2PP server port: " + e.toString(), 
        			Logger.DEBUG_P2P);
        	e.printStackTrace();
        	throw new RuntimeException("P2PEngine.initialize( ): Failed to open"
        			+ " the P2PP server port: " + e.toString());
        }   
        
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void asyncJoin() {

		if(this.dht.getConfigData().getBooleanProperty(
				ConfigData.BOOTSTRAP_MODE))
		{
			/* The bootstrap peer does not need to join */
			this.callbackListener.joinReady(
					new P2P4M2MCallbackResult(P2P4M2MCallbackResult.SUCCESS));
			return;
		}
		else {
			try {
				DHTEvent event = dht.join();
				if(event.getType() != DHTEvent.JOIN_PENDING) {
					this.callbackListener.joinReady(
							new P2P4M2MCallbackResult(
									P2P4M2MCallbackResult.SUCCESS));
				}
			}
			catch(Exception e) {
				this.logger.error("P2P4M2MEngine.asyncJoin( ): An Exception "
						+ "was caught: " + e.toString());
				e.printStackTrace();
				this.callbackListener.joinReady(new P2P4M2MCallbackResult(
						P2P4M2MCallbackResult.FAILED));
			}
		}
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void asyncLeave() {
		
		try {
			DHTEvent event = this.dht.leave();
		
			if(event.getType() != DHTEvent.LEAVE_PENDING) {
				logger.debug("P2P4M2MEngine.asyncLeave( ): returning "
						+ "immediately with result: " + event.toString());
				this.callbackListener.leaveReady(new P2P4M2MCallbackResult(
						P2P4M2MCallbackResult.SUCCESS));
			}
		}
		catch(Exception e) {
			logger.error("P2P4M2MEngine.asyncLeave( ): An Exception was caught:"
					+ " " + e.toString());
			e.printStackTrace();
			this.callbackListener.leaveReady(new P2P4M2MCallbackResult(
					P2P4M2MCallbackResult.FAILED));
		}
		
	}
		
	/**
	 * @inheritDoc
	 */
	public void asyncLookup(String lookupID, String key) 
	{
		byte[] resID = new byte[P2ppConstants.SHA_1_HASH_LENGTH];
		byte[] keyBytes = key.getBytes();
		SecureHashAlgorithm.computeMessageDigest(keyBytes, keyBytes.length, 
				resID);
		try {
			this.dht.lookup(resID, lookupID, ResourceRecord.BYTE_SEQUENCE_KIND, 
					this);
		}
		catch(Exception e) {
			logger.error("P2P4M2MEngine.asyncLookup( ): An Exception was "
					+ "caught: " + e.toString());
			e.printStackTrace();
		}
	}
	
	/**
	 * @inheritDoc
	 */
	public void asyncLookup(String lookupID, byte[] key) 
	{
		byte[] resID = new byte[P2ppConstants.SHA_1_HASH_LENGTH];
		byte[] keyBytes = key;
		SecureHashAlgorithm.computeMessageDigest(keyBytes, keyBytes.length, 
				resID);
		try {
			this.dht.lookup(resID, lookupID, ResourceRecord.BYTE_SEQUENCE_KIND, 
					this);
		}
		catch(Exception e) {
			logger.error("P2P4M2MEngine.asyncLookup( ): An Exception was "
					+ "caught: " + e.toString());
			e.printStackTrace();
		}
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void asyncStore(String storeID, String key, byte[] value) {
		byte[] resID = new byte[P2ppConstants.SHA_1_HASH_LENGTH];
		byte[] keyBytes = key.getBytes();
		SecureHashAlgorithm.computeMessageDigest(keyBytes, keyBytes.length, 
				resID);
		try {
			this.dht.publish(resID, value, -1, 
					ResourceRecord.BYTE_SEQUENCE_KIND, storeID, this);
		}
		catch(Exception e) {
			logger.error("P2P4M2MEngine.asyncStore( ): An Exception was caught:"
					+ " " + e.toString());
			e.printStackTrace();
		}
	}
	
	/**
	 * @inheritDoc
	 */
	public void asyncStore(String storeID, byte[] key, byte[] value) {
		byte[] resID = new byte[P2ppConstants.SHA_1_HASH_LENGTH];
		byte[] keyBytes = key;
		SecureHashAlgorithm.computeMessageDigest(keyBytes, keyBytes.length, 
				resID);
		try {
			this.dht.publish(resID, value, -1, 
					ResourceRecord.BYTE_SEQUENCE_KIND, storeID, this);
		}
		catch(Exception e) {
			logger.error("P2P4M2MEngine.asyncStore( ): An Exception was caught:"
					+ " " + e.toString());
			e.printStackTrace();
		}
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void asyncRemove(String removeID, String key) {
		
		byte[] resID = new byte[P2ppConstants.SHA_1_HASH_LENGTH];
		byte[] keyBytes = key.getBytes();
		SecureHashAlgorithm.computeMessageDigest(keyBytes, keyBytes.length, 
				resID);
		try {
			this.dht.remove(removeID, resID,  
					ResourceRecord.BYTE_SEQUENCE_KIND, this);
		}
		catch(Exception e) {
			logger.error("P2P4M2MEngine.asyncRemove( ): An Exception was caught:"
					+ " " + e.toString());
			e.printStackTrace();
		}
	}
	
	/**
	 * @inheritDoc
	 */
	public void asyncRemove(String removeID, byte[] key) {
		
		byte[] resID = new byte[P2ppConstants.SHA_1_HASH_LENGTH];
		byte[] keyBytes = key;
		SecureHashAlgorithm.computeMessageDigest(keyBytes, keyBytes.length, 
				resID);
		try {
			this.dht.remove(removeID, resID,  
					ResourceRecord.BYTE_SEQUENCE_KIND, this);
		}
		catch(Exception e) {
			logger.error("P2P4M2MEngine.asyncRemove( ): An Exception was caught:"
					+ " " + e.toString());
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Show text both in the log and in the console
	 */
	public void showText(String text, int logLevel, String caller) {
		logger.log(text, logLevel);
		logger.logToConsole(text, logLevel, caller);
	}
	
	
	/**
	 * Get the logger
	 */
	public P2pSipLogger getLogger() {
		return this.logger;
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void eventOccurred(DHTEvent event) {
		
		int eventCode = event.getType();
		
		if(eventCode == DHTEvent.JOIN_COMPLETE) {
			logger.debug("P2P4M2MEngine.eventOccurred( ): Join complete!");
			this.callbackListener.joinReady(new P2P4M2MCallbackResult(
					P2P4M2MCallbackResult.SUCCESS));
		}
		else if (event.isLeaveRelated()) {
			logger.debug("P2P4M2MEngine.eventOccurred( ): Leave complete!");
			this.callbackListener.leaveReady(new P2P4M2MCallbackResult(
					P2P4M2MCallbackResult.SUCCESS));			
		}
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void eventOccurred(int code) {
		DHTEvent event = new DHTEvent(code);
		eventOccurred(event);
	}
	

	/**
	 * @inheritDoc
	 */
	public void eventOccurred(LookupResult lookupEvent) {
		logger.error("P2P4M2MEngine.eventOccurred(LookupResult): Implementation"
				+ " missing");
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void infoToApplication(String message) {
		logger.debug("P2P4M2MEngine.infoToApplication( ): '" + message + "'");
	}
	
	
	/**
	 * @inheritDoc
	 */
	public MIDlet getMidlet() {
		return null;
	}
	
	
	/**
	 * @inheritDoc
	 */
	public byte[] tunneledDataReceived(byte[] data, byte[] senderID, 
			byte[] targetID) 
	{
		logger.error("P2P4M2MEngine.tunnelResultReceived( ): Implementation "
				+ "missing");
		return null;
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void incomingAppAttach(SipICEConnection conn) {
		logger.error("P2P4M2MEngine.incomingAppAttach( ): Implementation "
				+ "missing");		
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void candidateGatheringForIncomingAppAttachFinished(int delay) {
		logger.error("P2P4M2MEngine.candidateGatheringForIncomingAppAttach"
				+ "Finished( ): Implementation missing");
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void receiveLookupResult(LookupResult result) {
		
		int status = P2P4M2MLookupResult.SUCCESS;
		
		if(!result.isSuccessful()) {
			status = P2P4M2MLookupResult.FAILURE;
		}
		
		P2P4M2MLookupResult lr = new P2P4M2MLookupResult(status, 
				result.getData(), result.getLookupID());
		this.callbackListener.lookupReady(lr);
		
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void receivePublishResult(String id, int result, int responseCode) {
		
		int status = P2P4M2MCallbackResult.SUCCESS;
		
		if(status == IPublishResultReceiver.FAILED) {
			status = P2P4M2MCallbackResult.FAILED;
		}
		
		this.callbackListener.storeReady(id, new P2P4M2MCallbackResult(status));
		
	}
	
	
	/**
	 * @inheritDoc
	 */
	public void receiveRemoveResult(String id, int result, int responseCode) {
		int status = P2P4M2MCallbackResult.SUCCESS;
		
		if(status == IPublishResultReceiver.FAILED) {
			status = P2P4M2MCallbackResult.FAILED;
		}
		
		this.callbackListener.removeReady(id,
				new P2P4M2MCallbackResult(status));
		
	}
	
	
	/**
	 * Get the config data
	 * @return
	 */
	public ConfigData getConfig() {
		return this.dht.getConfigData();
	}
	
	/**
	 * Get the PeerID data
	 * @return
	 */
	public  byte[] getNodeID() {
		return this.dht.getChordData().getLocalPeerData().getPeerID().getBytes();
	}
	
	/**
	 * Get the IPEndpoint data
	 * @return 
	 */
	public byte[] getIP() {
		byte[] ipbytes = this.dht.getChordData().getConfigData().getLocalIP().getAddressBytes();
		byte[] ipfinal = new byte[16];
		if (ipbytes.length == 4) {
			System.arraycopy(ipbytes, 0, ipfinal, 0, ipbytes.length);
		}
		else {
			ipfinal = ipbytes;
		}
		
		return ipfinal;
	}
	
	public  IPEndpoint getIPEndpoint() {
		return this.dht.getChordData().getConfigData().getLocalIP();
	}
	/**
	 * Get the UserName data
	 * @return 
	 */
	public String getName() {
		return this.dht.getChordData().getConfigData().getProperty("userName");
	}

	
}
